"use strict";
var _ = require("lodash");
var path = require("path");
var stack_trace_1 = require("stack-trace");
var md5_1 = require("ts-md5/dist/md5");
var domain_1 = require("../domain");
var io_1 = require("../io");
var serenity_1 = require("../serenity");
var rehearsal_report_1 = require("./rehearsal_report");
function serenityBDDReporter(pathToReports) {
    if (pathToReports === void 0) { pathToReports = serenity_1.Default_Path_To_Reports; }
    return new SerenityBDDReporter(new io_1.FileSystem(pathToReports));
}
exports.serenityBDDReporter = serenityBDDReporter;
var SerenityBDDReporter = (function () {
    function SerenityBDDReporter(fs) {
        this.fs = fs;
    }
    SerenityBDDReporter.prototype.assignTo = function (stage) {
        this.stage = stage;
        this.stage.manager.registerInterestIn(SerenityBDDReporter.Events_of_Interest, this);
    };
    SerenityBDDReporter.prototype.notifyOf = function (event) {
        switch (event.constructor.name) {
            case domain_1.SceneFinished.name: return this.persistReportFor(event);
        }
    };
    SerenityBDDReporter.prototype.persistReportFor = function (_a) {
        var _this = this;
        var value = _a.value;
        var filename = md5_1.Md5.hashStr(value.subject.id) + ".json";
        this.stage.manager.informOfWorkInProgress(rehearsal_report_1.RehearsalReport.from(this.stage.manager.readNewJournalEntriesAs('SerenityBDDReporter'))
            .exportedUsing(new SerenityBDDReportExporter())
            .then(function (fullReport) { return Promise.all(fullReport.scenes.map(function (scene) { return _this.fs.store(filename, JSON.stringify(scene)); })); }));
    };
    return SerenityBDDReporter;
}());
SerenityBDDReporter.Events_of_Interest = [domain_1.SceneFinished];
exports.SerenityBDDReporter = SerenityBDDReporter;
/**
 * Transforms the tree structure of the RehearsalPeriod to a format acceptable by Protractor
 */
var SerenityBDDReportExporter = (function () {
    function SerenityBDDReportExporter() {
        this.errorExporter = new ErrorExporter();
        this.photoExporter = new PhotoExporter();
        this.dashified = function (name) { return name
            .replace(/([a-z])([A-Z])/g, '$1-$2')
            .replace(/[ \t\W]/g, '-')
            .replace(/^-+|-+$/g, '')
            .toLowerCase(); };
        this.isAnIssue = function (tag) { return !!~['issue', 'issues'].indexOf(tag.type); };
    }
    SerenityBDDReportExporter.prototype.exportRehearsal = function (node) {
        var _this = this;
        return Promise.all(node.children.map(function (child) { return child.exportedUsing(_this); }))
            .then(function (children) { return ({
            scenes: children,
        }); });
    };
    SerenityBDDReportExporter.prototype.exportScene = function (node) {
        var _this = this;
        return Promise.all(node.children.map(function (child) { return child.exportedUsing(_this); }))
            .then(function (children) { return ({
            title: node.value.name,
            name: node.value.name,
            description: '',
            startTime: node.startedAt,
            duration: node.duration(),
            driver: 'unknown',
            testSource: 'cucumber',
            manual: false,
            result: domain_1.Result[node.outcome.result],
            userStory: {
                id: _this.dashified(node.value.category),
                path: path.relative(process.cwd(), node.value.path),
                storyName: node.value.category,
                type: 'feature',
            },
            tags: _this.tagsFor(node.value),
            issues: _this.issuesCoveredBy(node.value),
            testSteps: children,
            annotatedResult: domain_1.Result[node.outcome.result],
            testFailureCause: _this.errorExporter.tryToExport(node.outcome.error),
        }); });
    };
    SerenityBDDReportExporter.prototype.exportActivity = function (node) {
        var _this = this;
        return Promise.all(node.children.map(function (child) { return child.exportedUsing(_this); }))
            .then(function (children) { return node.photos().then(function (photos) { return ({
            description: node.value.name,
            duration: node.duration(),
            startTime: node.startedAt,
            screenshots: _this.photoExporter.tryToExport(photos),
            result: domain_1.Result[node.outcome.result],
            children: children,
            exception: _this.errorExporter.tryToExport(node.outcome.error),
        }); }); });
    };
    SerenityBDDReportExporter.prototype.issuesCoveredBy = function (scene) {
        var onlyIssueTags = this.isAnIssue, toIssueIds = function (tag) { return tag.values; };
        return _.chain(scene.tags).filter(onlyIssueTags).map(toIssueIds).flatten().uniq().value();
    };
    SerenityBDDReportExporter.prototype.tagsFor = function (scene) {
        var isAnIssue = this.isAnIssue;
        function serialise(tag) {
            var noValue = function (t) { return ({ name: t.type, type: 'tag' }); }, withValue = function (t) { return ({ name: t.values.join(','), type: t.type }); };
            return tag.values.length === 0
                ? noValue(tag)
                : withValue(tag);
        }
        function breakDownIssues(tag) {
            return isAnIssue(tag)
                ? tag.values.map(function (issueId) { return new domain_1.Tag('issue', [issueId]); })
                : tag;
        }
        function featureTag(featureName) {
            return {
                name: featureName,
                type: 'feature',
            };
        }
        return _.chain(scene.tags)
            .map(breakDownIssues)
            .flatten()
            .map(serialise)
            .uniqBy('name')
            .concat(featureTag(scene.category))
            .value();
    };
    return SerenityBDDReportExporter;
}());
exports.SerenityBDDReportExporter = SerenityBDDReportExporter;
var PhotoExporter = (function () {
    function PhotoExporter() {
        this.ifNotEmpty = function (list) { return !!list.length ? list : undefined; };
    }
    PhotoExporter.prototype.tryToExport = function (photos) {
        return this.ifNotEmpty(photos.map(function (photo) { return ({ screenshot: path.basename(photo.path) }); }));
    };
    return PhotoExporter;
}());
var ErrorExporter = (function () {
    function ErrorExporter() {
    }
    ErrorExporter.prototype.tryToExport = function (error) {
        if (!error) {
            return undefined; // an undefined JSON field does not get serialised and that's what Serenity BDD expects
        }
        return {
            errorType: error.name,
            message: error.message,
            stackTrace: this.stackTraceOf(error),
        };
    };
    ErrorExporter.prototype.stackTraceOf = function (error) {
        var serenityCode = /node_modules[\\/]serenity/, onlyIfFound = function (index) { return !!~index ? index : undefined; }, firstSerenityStackFrame = function (stack) { return onlyIfFound(stack.findIndex(function (frame) { return !!serenityCode.exec(frame.getFileName()); })); }, parsed = stack_trace_1.parse(error);
        return parsed.slice(0, firstSerenityStackFrame(parsed)).map(function (frame) {
            return {
                declaringClass: frame.getTypeName() || frame.getFunctionName() || '',
                methodName: frame.getMethodName() || frame.getFunctionName() || '',
                fileName: frame.getFileName(),
                lineNumber: frame.getLineNumber(),
            };
        });
    };
    return ErrorExporter;
}());
//# sourceMappingURL=serenity_bdd_reporter.js.map